"""Search/Retrieval via URL for Ex Libris' Alma

This module is intended for searching the library software system Alma
for single or multiple records. The search for single records must be
done with a unique ID (e. g. Barcode) and will return exactly one
xml.etree.ElementTree Element. The search for multiple records will
return an iterable containing such Elements.

When querying for multiple records, keep in mind that Ex Libris limits
the retrieval of records to 10k.

To use this module you will need to know at least the following
information:
* The subdomain of the Alma installation
* The institution code of the institution you want to query
* Which format you want returned

Possible formats are documented here:
https://developers.exlibrisgroup.com/alma/integrations/SRU/

The Library of Congrass has general information on SRU:
http://www.loc.gov/standards/sru/
"""

from urllib import request, parse
from xml.etree.ElementTree import fromstring, Element


class SruException(Exception):
    pass


class NoRecord(SruException):
    pass


class MultiRecord(SruException):
    pass


class SruDiagnostics(SruException):
    pass


class RecordRetriever:
    """Class to initialize searchRetrieve-Requests.

    Some standard-methods for fetching single records with a unique
    identifier:
    """

    def __init__(self,
                 subdomain: str,
                 inst_code: str,
                 xml_format: str):
        """
        :param subdomain: Subdomain (preferably CNAME) for the Alma Instance
        :param inst_code: Alma Institution Code
        :param xml_format: Format of the returned xml
        """

        self.url_base = ('https://'
                         f'{subdomain}.alma.exlibrisgroup.com/'
                         f'view/sru/{inst_code}?')
        self.url_query = {'version': '1.2',
                          'operation': 'searchRetrieve',
                          'recordSchema': xml_format}

    def get_single(self, alma_query: str) -> Element:
        """Return one MARC XML record for a specific query.

        :param alma_query: URI-parameter query for SRU request
        :return: XML as etree-Element
        :rtype Element
        :raises NoRecord: if no record was returned by SRU
        :raises Multirecord: if > 1 record was returned for uniq ID
        :raises SruDiagnostics: if SRU returns an error
        """

        self.url_query['startRecord'] = '0'
        self.url_query['maximumRecords'] = '1'
        self.url_query['query'] = alma_query
        url_query_string = parse.urlencode(self.url_query)
        sru_url = self.url_base + url_query_string

        response = request.urlopen(sru_url).read()
        sru_xml = fromstring(response)

        namespaces = {
            'srw': 'http://www.loc.gov/zing/srw/',
            'diag': 'http://www.loc.gov/zing/srw/diagnostic/',
            'marc': 'http://www.loc.gov/MARC21/slim'
        }

        num_recs = sru_xml.find("srw:numberOfRecords", namespaces)

        if (
            num_recs is not None and
            num_recs.text == '0'
        ):
            raise NoRecord('No matching records found!')

        elif (
            num_recs is not None and
            num_recs.text > '1'
        ):
            raise MultiRecord('More than one record found'
                              'while exactly one was expected!')

        elif sru_xml.find("srw:diagnostics", namespaces):
            xpath = "srw:diagnostics/diag:diagnostic/diag:message"
            sru_message = sru_xml.find(xpath, namespaces).text
            raise SruDiagnostics(sru_message)

        else:
            record_xpath = 'srw:records/srw:record//marc:record'
            marc_xml = sru_xml.find(record_xpath, namespaces)
            return marc_xml

    def by_marc_009(self, marc_009: str) -> Element:
        """Return record by MARC 009 (control number).

        :param marc_009: MARC bibliographic record control number.
        """

        alma_query = f'alma.local_control_field_009={marc_009}'
        return self.get_single(alma_query)

    def by_barcode(self, barcode: str) -> Element:
        """Return record by barcode.

        param barcode: Barcode of a specific item."""

        alma_query = f'alma.barcode={barcode}'
        return self.get_single(alma_query)

    def by_mms_id(self, mms_id: str) -> Element:
        """Return record by MMS-ID.

        param mmsid: MMS-ID of a specific item."""

        alma_query = f'alma.mms_id={mms_id}'
        return self.get_single(alma_query)
